"""Test tools module"""

"""
Copyright (c) 2018, EPFL/Blue Brain Project

 This file is part of BluePyMM <https://github.com/BlueBrain/BluePyMM>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import os
import pandas
from string import digits

import pytest

from bluepymm import tools

BASE_DIR = os.path.abspath(os.path.dirname(__file__))
EXAMPLES = os.path.join(BASE_DIR, 'examples')
TMP_DIR = os.path.join(BASE_DIR, 'tmp/tools')


@pytest.mark.unit
def test_cd():
    """bluepymm.tools: test cd"""

    old_cwd = os.getcwd()
    with tools.cd(EXAMPLES):
        assert os.getcwd() == EXAMPLES

    assert old_cwd == os.getcwd()


@pytest.mark.unit
def test_json():
    """bluepymm.tools: test load_json and write_json"""
    output_dir = TMP_DIR
    output_name = 'tmp.json'
    config = {'test': ['1', 'two']}

    tools.makedirs(output_dir)
    ret_path = tools.write_json(output_dir, output_name, config)
    assert os.path.join(output_dir, output_name) == ret_path
    ret = tools.load_json(ret_path)
    assert config == ret


@pytest.mark.unit
def test_makedirs():
    """bluepymm.tools: test makedirs"""
    make_dir = os.path.join(TMP_DIR, 'make_dir')
    tools.makedirs(make_dir)
    assert os.path.isdir(make_dir)

    # try again -> no error
    make_dir = os.path.join(TMP_DIR, 'make_dir')
    tools.makedirs(make_dir)
    assert os.path.isdir(make_dir)

    # causes error that is not caught
    with pytest.raises(OSError):
        tools.makedirs('')


@pytest.mark.unit
def test_check_no_null_nan_values():
    """bluepymm.tools: test check_no_null_nan_values"""
    data = pandas.DataFrame([[1, 2], [3, 4]], columns=list('AB'))
    assert tools.check_no_null_nan_values(data, 'test')


@pytest.mark.unit
def test_check_no_null_nan_values_nan():
    """bluepymm.tools: test check_no_null_nan_values with nan"""
    data = pandas.DataFrame([[1, float('nan')], [3, 4]], columns=list('AB'))
    with pytest.raises(Exception):
        tools.check_no_null_nan_values(data, 'test')


@pytest.mark.unit
def test_check_no_null_nan_values_none():
    """bluepymm.tools: test check_no_null_nan_values with None"""
    data = pandas.DataFrame([[1, 2], [None, 4]], columns=list('AB'))
    with pytest.raises(Exception):
        tools.check_no_null_nan_values(data, 'test')


@pytest.mark.unit
def test_check_all_combos_have_run():
    """bluepymm.tools: test check_all_combos_have_run"""
    data = pandas.DataFrame({'to_run': [False, False, False],
                             'field': [1, 2, 3]})
    assert tools.check_all_combos_have_run(data, 'test')

    data = pandas.DataFrame({'to_run': [True, True, True],
                             'field': [1, 2, 3]})
    with pytest.raises(Exception):
        tools.check_all_combos_have_run(data, 'test')

    data = pandas.DataFrame({'to_run': [False, True, False],
                             'field': [1, 2, 3]})
    with pytest.raises(Exception):
        tools.check_all_combos_have_run(data, 'test')


@pytest.mark.unit
def test_load_module():
    """bluepymm.tools: test load_module"""
    # load module
    module_dir = os.path.join(EXAMPLES, 'simple1/data/emodels_dir/subdir/')
    setup = tools.load_module('setup', module_dir)
    # try and execute something from loaded module
    setup.evaluator.create('emodel1')

    # load as file
    setup_dir = os.path.join(module_dir, 'setup')
    evaluator = tools.load_module('evaluator', setup_dir)
    evaluator.create('emodel1')


@pytest.mark.unit
def test_check_compliance_with_neuron():
    """bluepymm.tools: test check compliance with neuron template name rules"""
    not_compl = ['', '1test', 'test-test',
                 'testtesttesttesttesttesttesttesttesttesttesttesttesttesttes']
    for name in not_compl:
        assert not tools.check_compliance_with_neuron(name)

    compliant = ['test_tesT', 'test123test', 'Test']
    for name in compliant:
        assert tools.check_compliance_with_neuron(name)


@pytest.mark.unit
def test_shorten_and_hash_string():
    """bluepymm.tools: test convert string"""
    label = 'testtesttesttesttesttesttesttesttest'
    assert label == tools.shorten_and_hash_string(label)

    keep_length = 3
    hash_length = 20
    expected_length = keep_length + hash_length + 1
    ret = tools.shorten_and_hash_string(label, keep_length=keep_length,
                                        hash_length=hash_length)
    assert label != ret
    assert len(ret) == expected_length
    assert label[0:keep_length] == ret[0:keep_length]
    assert '_' == ret[keep_length]

    hash_length = 21
    with pytest.raises(ValueError):
        tools.shorten_and_hash_string(label, keep_length, hash_length)


@pytest.mark.unit
def test_get_neuron_compliant_template_name():
    """bluepymm.tools: test get neuron-compliant template name"""
    name = 'test'
    assert tools.check_compliance_with_neuron(name)
    ret = tools.get_neuron_compliant_template_name(name)
    assert ret == name
    assert tools.check_compliance_with_neuron(ret)

    name = '123test-test'
    assert not tools.check_compliance_with_neuron(name)
    ret = tools.get_neuron_compliant_template_name(name)
    assert ret == name.lstrip(digits).replace('-', '_')
    assert tools.check_compliance_with_neuron(ret)

    name = 'testtesttesttesttesttesttesttesttesttesttesttesttesttesttest'
    assert not tools.check_compliance_with_neuron(name)
    ret = tools.get_neuron_compliant_template_name(name)
    assert tools.check_compliance_with_neuron(ret)


@pytest.mark.unit
def test_decode_bstring():
    """bluepymm.tools test the bstring decoding function."""
    bstr_obj = b"this is a byte string"
    decoded_bstr = "this is a byte string"
    assert tools.decode_bstring(bstr_obj) == decoded_bstr

    str_obj = "this is a string"
    assert tools.decode_bstring(str_obj) == str_obj



"""Test bluepymm/run_combos"""

"""
Copyright (c) 2018, EPFL/Blue Brain Project

 This file is part of BluePyMM <https://github.com/BlueBrain/BluePyMM>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import os
import shutil

from bluepymm import tools, run_combos


BASE_DIR = os.path.abspath(os.path.dirname(__file__))
TEST_DATA_DIR = os.path.join(BASE_DIR, 'examples/simple1')
TMP_DIR = os.path.join(BASE_DIR, 'tmp/run_combos')


def _verify_run_combos_output(scores_db):
    """Helper function to verify output run combos"""
    # TODO: test database contents
    assert os.path.isfile(scores_db)


def test_run_combos():
    """bluepymm.run_combos: test run_combos based on example simple1"""
    config_template_path = 'simple1_conf_run.json'

    with tools.cd(TEST_DATA_DIR):
        # prepare input data
        shutil.copytree('output_expected', TMP_DIR)
        config = tools.load_json(config_template_path)
        config['scores_db'] = os.path.join(TMP_DIR, 'scores.sqlite')
        config['output_dir'] = TMP_DIR

        # run combination preparation
        run_combos.main.run_combos_from_conf(config)

        # verify output
        _verify_run_combos_output(config['scores_db'])

        # with use_apical_points
        config['use_apical_points'] = False
        run_combos.main.run_combos_from_conf(config)
        _verify_run_combos_output(config['scores_db'])



"""Python Model Management"""

"""
Copyright (c) 2018, EPFL/Blue Brain Project

 This file is part of BluePyMM <https://github.com/BlueBrain/BluePyMM>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


# pylint: disable=C0325, W0223, R0914, E1121, E1123

import sys
import os
import sh
import shutil
import traceback
import multiprocessing
import tarfile

from bluepymm import tools


def check_emodels_in_repo(conf_dict):
    """Check whether input e-models are organized in branches of a repository.

    Args:
        conf_dict: A dict with either the key "emodels_repo" or the key
            "emodels_dir".

    Returns:
        True if the input e-models are organized in separate branches of a
        git repository, false if the e-models are organized into separate
        subdirectories.

    Raises:
        ValueError: if both or none of the keys "emodels_repo" and
            "emodels_dir" are present.

    TODO: replace "emodels_repo" and "emodels_dir" by "emodels_input_type" and
        "emodels_path" or similar.
    """
    if 'emodels_repo' in conf_dict and 'emodels_dir' in conf_dict:
        raise ValueError("Impossible to specify both 'emodels_repo' and"
                         " 'emodels_dir' in configuration file")
    elif 'emodels_repo' in conf_dict:
        emodels_in_repo = True
    elif 'emodels_dir' in conf_dict:
        emodels_in_repo = False
    else:
        raise ValueError("Need to specify either 'emodels_dir' or"
                         " 'emodels_repo' in configuration file")
    return emodels_in_repo


def convert_emodel_input(emodels_in_repo, conf_dict, continu):
    """Convert e-model input to BluePyMM file structure and return path to that
    structure.

    Args:
        emodels_in_repo: True if the input e-models are organized in separate
            branches of a git repository, false if the e-models are organized
            into separate subdirectories.
        conf_dict: A dict with e-model input configuration.
        continu: True if this BluePyMM run builds on a previous run, False
            otherwise

    Returns:
        Path to BluePyMM file structure.
    """
    tmp_emodels_dir = os.path.abspath(os.path.join(conf_dict['tmp_dir'],
                                                   'emodels_repo'))
    if not continu:
        if emodels_in_repo:
            print('Cloning input e-models repository in %s' % tmp_emodels_dir)
            sh.git('clone', conf_dict['emodels_repo'], tmp_emodels_dir)

            with tools.cd(tmp_emodels_dir):
                sh.git('checkout', conf_dict['emodels_githash'])
        else:
            if (os.path.exists(tmp_emodels_dir)
               and os.path.isdir(tmp_emodels_dir)):
                shutil.rmtree(tmp_emodels_dir)
            shutil.copytree(conf_dict['emodels_dir'], tmp_emodels_dir)
    return tmp_emodels_dir


def get_emodel_dicts(emodels_dir, final_json_path, emodel_etype_map_path):
    """Read and return detailed e-model information.

    Args:
        emodels_dir: Path to BluePyMM file structure.
        final_json_path: Path to final e-model map, relative to
            `emodels_dir`.
        emodel_etype_map_path: Path to e-model e-type map, relative to
            `emodels_dir`.

    Returns:
        (string, dict, dict)-tuple with:
            - final e-model map,
            - e-model e-type map,
            - name of directory containing final e-model map.
    """
    final_dict_path = os.path.join(emodels_dir, final_json_path)
    final_dict = tools.load_json(final_dict_path)
    e_map_path = os.path.join(emodels_dir, emodel_etype_map_path)
    emodel_etype_map = tools.load_json(e_map_path)
    dict_dir = os.path.dirname(final_dict_path)
    return final_dict, emodel_etype_map, dict_dir


def create_and_write_hoc_file(emodel, emodel_dir, hoc_dir, emodel_params,
                              template, morph_path=None,
                              model_name=None):
    """Create .hoc code for a given e-model based on code from
    '<emodel_dir>/setup', e-model parameters and a given template, and write
    out the result to a file named <hoc_dir>/<model_name or emodel>.hoc.

    Args:
        emodel: e-model name
        emodel_dir: the directory containing a module 'setup', which describes
                    the e-model
        hoc_dir: the directory to which the resulting .hoc file will be written
                 out.
        emodel_params: a dict with e-model parameters
        template: template file used for the creation of the .hoc file
        morph_path: path to morphology file, used to overwrite the original
                    morphology of an e-model. Default is None.
        model_name: used to name the .hoc file. If None, the e-model name is
                    used. Default is None.
    """
    setup = tools.load_module('setup', emodel_dir)

    with open(os.devnull, 'w') as devnull:
        old_stdout = sys.stdout
        try:
            sys.stdout = devnull
            if hasattr(setup, 'multieval'):
                evaluator = setup.evaluator.create(emodel).evaluators[0]
            else:
                evaluator = setup.evaluator.create(emodel)

            # set path to morphology
            if morph_path is not None:
                evaluator.cell_model.morphology.morphology_path = morph_path

            # set template name
            template_name = model_name or emodel
            evaluator.cell_model.name = template_name
            evaluator.cell_model.check_name()
        finally:
            sys.stdout = old_stdout

    # create hoc code
    template_dir = os.path.dirname(template)

    for mechanism in evaluator.cell_model.mechanisms:
        if 'Stoch' in mechanism.prefix:
            mechanism.deterministic = False
    hoc = evaluator.cell_model.create_hoc(emodel_params, template=template,
                                          template_dir=template_dir)

    # write out result
    hoc_file_name = '{}.hoc'.format(template_name)
    emodel_hoc_path = os.path.join(hoc_dir, hoc_file_name)
    with open(emodel_hoc_path, 'w') as emodel_hoc_file:
        emodel_hoc_file.write(hoc)


def prepare_emodel_dir(input_args):
    """Clone e-model input and prepare the e-model directory.

    Args:
        input_args: 9-tuple
            - original_emodel(str): e-model name
            - emodel(str): e-model name
            - emodel_dict: dict with all e-model parameters
            - emodels_dir: directory with all e-models
            - opt_dir: directory with all opt e-models (TODO: clarify)
            - hoc_dir: absolute path to the directory to which the .hoc
            files will be written out
            - hoc_template: path to the jinja hoc template
            - emodels_in_repo: True if the input e-models are organized
            in separate branches of
            a git repository, false if the e-models are organized into
            separate subdirectories.
            - continu: True if this BluePyMM run builds on a previous run,
            False otherwise

    Returns:
        A dict mapping the e-model and the original e-model to the e-model dir
    """
    original_emodel, emodel, emodel_dict, emodels_dir, \
        opt_dir, hoc_dir, hoc_template, emodels_in_repo, continu = input_args

    try:
        print('Preparing: %s' % emodel)
        emodel_dir = os.path.join(emodels_dir, emodel)

        if not continu:
            tar_filename = os.path.abspath(
                os.path.join(
                    emodels_dir,
                    '%s.tar' %
                    emodel))

            main_path = emodel_dict.get('main_path', '.')

            with tools.cd(os.path.join(opt_dir, main_path)):
                if emodels_in_repo:
                    sh.git(
                        'archive',
                        '--format=tar',
                        '--prefix=%s/' % emodel,
                        'origin/%s' % emodel_dict['branch'],
                        _out=tar_filename)
                else:
                    with tarfile.open(tar_filename, 'w') as tar_file:
                        tar_file.add('.', arcname=emodel)

            # extract in e-model directory, compile mechanisms and create .hoc
            # TODO: clean up .tar file?
            with tools.cd(emodels_dir):
                sh.tar('xf', tar_filename)

                with tools.cd(emodel):
                    print('Compiling mechanisms ...')
                    if os.path.exists('mechanisms'):
                        sh.nrnivmodl('mechanisms')

                    create_and_write_hoc_file(
                        emodel, emodel_dir, hoc_dir, emodel_dict['params'],
                        template=hoc_template)

                os.remove(tar_filename)

    except Exception:
        raise Exception(''.join(traceback.format_exception(*sys.exc_info())))

    return {emodel: emodel_dir, original_emodel: emodel_dir}


def prepare_emodel_dirs(
    final_dict,
    emodel_etype_map,
    emodels_dir,
    opt_dir,
    emodels_hoc_dir,
    emodels_in_repo,
    hoc_template,
    continu=False,
    n_processes=None,
):
    """Prepare the directories for the emodels.

    Args:
        final_dict: final e-model map
        emodel_etype_map: e-model e-type map
        emodels_dir: absolute path to the directory with all e-models. This
            directory is created by this function if it does not exist yet.
        opt_dir: directory with all opt e-models (TODO: clarify)
        emodels_hoc_dir: absolute path to the directory to which the .hoc files
            will be written out. Created by this function if it does not exist
            yet.
        emodels_in_repo: True if the input e-models are organized in separate
            branches of a git repository, false if the e-models are organized
            into separate subdirectories.
        hoc_template: path to the jinja hoc template.
        continu: True if this BluePyMM run builds on a previous run, False
            otherwise. Default is False.
        n_processes: the integer number of processes. If `None`,
        all processes are going to be used.

    Return:
        A dict mapping e-models to prepared e-model directories.
    """
    tools.makedirs(emodels_dir)
    tools.makedirs(emodels_hoc_dir)

    arg_list = []
    for original_emodel in emodel_etype_map:
        emodel = emodel_etype_map[original_emodel]['mm_recipe']
        emodel_dict = final_dict[original_emodel]
        arg_list.append(
            (original_emodel,
             emodel,
             emodel_dict,
             emodels_dir,
             opt_dir,
             emodels_hoc_dir,
             hoc_template,
             emodels_in_repo,
             continu))

    emodel_dirs = {}
    emodel_dir_dicts = []

    if n_processes == 1:
        for arg in arg_list:
            emodel_dir_dict = prepare_emodel_dir(arg)
            emodel_dir_dicts.append(emodel_dir_dict)
    else:
        print('Parallelising preparation of e-model directories')
        pool = multiprocessing.Pool(processes=n_processes,
                                    maxtasksperchild=1)
        for emodel_dir_dict in pool.map(prepare_emodel_dir, arg_list,
                                        chunksize=1):
            emodel_dir_dicts.append(emodel_dir_dict)

    for emodel_dir_dict in emodel_dir_dicts:
        for emodel, emodel_dir in emodel_dir_dict.items():
            emodel_dirs[emodel] = emodel_dir

    return emodel_dirs



"""Test bluepymm/select_combos"""

from __future__ import print_function

"""
Copyright (c) 2018, EPFL/Blue Brain Project

 This file is part of BluePyMM <https://github.com/BlueBrain/BluePyMM>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import os
import shutil
import filecmp

from bluepymm import tools, select_combos

BASE_DIR = os.path.abspath(os.path.dirname(__file__))
TEST_DATA_DIR = os.path.join(BASE_DIR, 'examples/simple1')
TMP_DIR = os.path.join(BASE_DIR, 'tmp/select_combos')


def _verify_output(benchmark_dir, output_dir):
    """Helper function to verify output of combination selection"""
    files = ['extneurondb.dat', 'mecombo_emodel.tsv']
    matches = filecmp.cmpfiles(benchmark_dir, output_dir, files)
    if len(matches[0]) != len(files):
        print('Mismatch in files: {}'.format(matches[1]))
    assert len(matches[0]) == len(files)
    assert os.path.exists(os.path.join(output_dir, 'mecombo_release.json'))


def _config_select_combos(config_template_path, tmp_dir):
    """Helper function to prepare input data for select_combos"""
    # copy input data
    shutil.copytree('output_expected', tmp_dir)

    # set configuration dict
    config = tools.load_json(config_template_path)
    config['scores_db'] = os.path.join(tmp_dir, 'scores.sqlite')
    config['pdf_filename'] = os.path.join(tmp_dir, 'megating.pdf')
    config['output_dir'] = os.path.join(tmp_dir, 'output')
    config['emodels_hoc_dir'] = os.path.join(tmp_dir, 'output/emodels_hoc')
    return config


def _test_select_combos(test_data_dir, tmp_dir, config_template_path,
                        benchmark_dir, n_processes=None):
    """Helper function to perform functional test of select_combos"""
    with tools.cd(test_data_dir):
        # prepare input data
        config = _config_select_combos(config_template_path, tmp_dir)

        # run combination selection
        select_combos.main.select_combos_from_conf(config, n_processes)

        # verify output
        _verify_output(benchmark_dir, config['output_dir'])


def test_select_combos():
    """bluepymm.select_combos: test select_combos based on example simple1"""
    config_template_path = 'simple1_conf_select.json'
    benchmark_dir = 'output_megate_expected'
    tmp_dir = os.path.join(TMP_DIR, 'test_select_combos')

    _test_select_combos(TEST_DATA_DIR, tmp_dir, config_template_path,
                        benchmark_dir, n_processes=1)


def test_select_combos_2():
    """bluepymm.select_combos: test select_combos based on example simple1 bis
    """
    config_template_path = 'simple1_conf_select_2.json'
    benchmark_dir = 'output_megate_expected'
    tmp_dir = os.path.join(TMP_DIR, 'test_select_combos_2')

    _test_select_combos(TEST_DATA_DIR, tmp_dir, config_template_path,
                        benchmark_dir)



"""Test prepare_emodel_dirs"""

"""
Copyright (c) 2018, EPFL/Blue Brain Project

 This file is part of BluePyMM <https://github.com/BlueBrain/BluePyMM>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import os
import shutil

import pytest

from bluepymm.prepare_combos import prepare_emodel_dirs
from bluepymm import tools


BASE_DIR = os.path.abspath(os.path.dirname(__file__))
TEST_DATA_DIR = os.path.join(BASE_DIR, 'examples/simple1')
TMP_DIR = os.path.join(BASE_DIR, 'tmp/test_prepare_emodel_dirs')
TEMPLATE_DIR = os.path.join(BASE_DIR, '../bluepymm/templates')


def teardown_module():
    """Remove the temporary files."""
    shutil.rmtree(TMP_DIR)


@pytest.mark.unit
def test_check_emodels_in_repo():
    """prepare_combos.prepare_emodel_dirs: test check_emodels_in_repo.
    """
    config = {'emodels_repo': 'test'}
    assert prepare_emodel_dirs.check_emodels_in_repo(config)

    config = {'emodels_dir': 'test'}
    assert not prepare_emodel_dirs.check_emodels_in_repo(config)

    config = {'emodels_dir': 'test', 'emodels_repo': 'test'}
    with pytest.raises(ValueError):
        prepare_emodel_dirs.check_emodels_in_repo(config)

    config = {}
    with pytest.raises(ValueError):
        prepare_emodel_dirs.check_emodels_in_repo(config)


def _test_convert_emodel_input(test_dir, emodels_in_repo, conf_dict, continu):
    """Helper function to test convert_emodel_input"""
    with tools.cd(test_dir):
        ret_dir = prepare_emodel_dirs.convert_emodel_input(emodels_in_repo,
                                                           conf_dict, continu)
        assert os.path.isdir(ret_dir)
        for k in ['emodel_etype_map_path', 'final_json_path']:
            assert os.path.isfile(os.path.join(ret_dir, conf_dict[k]))


@pytest.mark.unit
def test_convert_emodel_input_dir():
    """prepare_combos.prepare_emodel_dirs: test convert_emodel_input for dir
    based on test example 'simple1' with directory input.
    """
    conf_dict = {'emodels_dir': './data/emodels_dir',
                 'emodel_etype_map_path': 'subdir/emodel_etype_map.json',
                 'final_json_path': 'subdir/final.json',
                 'tmp_dir': os.path.join(
                     TMP_DIR, 'test_convert_emodel_input_dir'), }
    emodels_in_repo = False
    continu = False
    _test_convert_emodel_input(TEST_DATA_DIR, emodels_in_repo, conf_dict,
                               continu)
    # the second run is to make sure that the tmp_dir gets deleted and recopied
    # in the consecutive runs
    _test_convert_emodel_input(TEST_DATA_DIR, emodels_in_repo, conf_dict,
                               continu)

# TODO : how to do the test below?
# @pytest.mark.unit
# def test_convert_emodel_input_repo():
#     """prepare_combos.prepare_emodel_dirs: test convert_emodel_input for repo
#     based on test example 'simple1' with repository input.
#     """
#     conf_dict = {'emodels_repo': './tmp_git',
#                  'emodels_githash': 'master',
#                  'emodel_etype_map_path': 'subdir/emodel_etype_map.json',
#                  'final_json_path': 'subdir/final.json',
#                  'tmp_dir': os.path.join(
#                      TMP_DIR, 'test_convert_emodel_input_dir'), }
#                  }
#     emodels_in_repo = True
#     continu = False
#     _test_convert_emodel_input(TEST_DATA_DIR, emodels_in_repo, conf_dict,
#                                continu)


@pytest.mark.unit
def test_get_emodel_dicts():
    """prepare_combos.prepare_emodel_dirs: test get_emodel_dicts
    based on test example 'simple1'.
    """
    emodels_dir = './data/emodels_dir'
    final_json_path = 'subdir/final.json'
    emodel_etype_map_path = 'subdir/emodel_etype_map.json'

    expected_final_dict = {
        'emodel1': {
            'main_path': '.',
            'seed': 2,
            'rank': 0,
            'notes': '',
            'branch': 'emodel1',
            'params': {
                'cm': 1.0},
            'fitness': {
                'Step1.SpikeCount': 20.0},
            'score': 104.72906197480131,
            'morph_path': 'morphologies/morph1.asc'},
        'emodel2': {
            'main_path': '.',
            'seed': 2,
            'rank': 0,
            'notes': '',
            'branch': 'emodel2',
            'params': {
                'cm': 0.5},
            'fitness': {
                'Step1.SpikeCount': 20.0},
            'score': 104.72906197480131,
            'morph_path': 'morphologies/morph2.asc'}}
    expected_emodel_etype_map = {'emodel1': {'mm_recipe': 'emodel1',
                                             'etype': 'etype1',
                                             'layer': ['1', 'str1']},
                                 'emodel2': {'mm_recipe': 'emodel2',
                                             'etype': 'etype2',
                                             'layer': ['1', '2']}}
    expected_dict_dir = os.path.dirname(os.path.join(emodels_dir,
                                                     final_json_path))

    with tools.cd(TEST_DATA_DIR):
        final, ee_map, dict_dir = prepare_emodel_dirs.get_emodel_dicts(
            emodels_dir, final_json_path, emodel_etype_map_path)

    assert expected_final_dict == final
    assert expected_emodel_etype_map == ee_map
    assert expected_dict_dir == dict_dir


def _test_create_and_write_hoc_file(test_dir,
                                    emodel,
                                    emodel_dir,
                                    hoc_dir,
                                    emodel_parameters,
                                    template,
                                    morph_path,
                                    model_name):
    """Test create_and_write_hoc_files"""
    with tools.cd(test_dir):
        tools.makedirs(hoc_dir)

        prepare_emodel_dirs.create_and_write_hoc_file(
            emodel, emodel_dir, hoc_dir, emodel_parameters, template,
            morph_path=morph_path, model_name=model_name)

        # TODO: test hoc file contents
        template_name = model_name or emodel
        hoc_filename = '{}.hoc'.format(template_name)
        hoc_path = os.path.join(hoc_dir, hoc_filename)
        assert os.path.isfile(hoc_path)


@pytest.mark.unit
def test_create_and_write_hoc_file_none():
    """prepare_combos.prepare_emodel_dirs: test create_and_write_hoc_file
    based on morph1 of test example 'simple1'.
    """
    emodel = 'emodel1'
    emodel_dir = './data/emodels_dir/subdir/'
    hoc_dir = './output/emodels_hoc'
    emodel_parameters = {'cm': 1.0}
    template = 'cell_template_neuron.jinja2'
    morph_path = None
    model_name = None

    _test_create_and_write_hoc_file(TEST_DATA_DIR, emodel, emodel_dir, hoc_dir,
                                    emodel_parameters, template, morph_path,
                                    model_name)


@pytest.mark.unit
def test_create_and_write_hoc_file_morph_path_model_name():
    """prepare_combos.prepare_emodel_dirs: test create_and_write_hoc_file
    based on morph1 of test example 'simple1'.
    """
    emodel = 'emodel1'
    emodel_dir = './data/emodels_dir/subdir/'
    hoc_dir = './output/emodels_hoc'
    emodel_parameters = {'cm': 1.0}
    template = 'cell_template_neuron.jinja2'
    morph_path = 'morph.asc'
    model_name = 'test'

    _test_create_and_write_hoc_file(TEST_DATA_DIR, emodel, emodel_dir, hoc_dir,
                                    emodel_parameters, template, morph_path,
                                    model_name)


@pytest.mark.unit
def test_prepare_emodel_dir():
    """prepare_combos.prepare_emodel_dirs: test prepare_emodel_dir
    based on test example 'simple1'.
    """
    # create test directory, where output of this test will be created
    test_dir = os.path.join(TMP_DIR, 'test_prepare_emodel_dir')
    tools.makedirs(test_dir)

    # input parameters
    original_emodel = 'emodel1'
    emodel = 'emodel1'
    emodel_dict = {'main_path': '.',
                   'seed': 2,
                   'rank': 0,
                   'notes': '',
                   'branch': 'emodel1',
                   'params': {'cm': 1.0},
                   'fitness': {'Step1.SpikeCount': 20.0},
                   'score': 104.72906197480131,
                   'morph_path': 'morphologies/morph1.asc'}
    emodels_dir = os.path.join(test_dir, 'tmp/emodels/')
    opt_dir = os.path.join(TEST_DATA_DIR, 'data/emodels_dir/subdir/')
    hoc_dir = os.path.join(test_dir, 'output/emodels_hoc/')
    hoc_template = os.path.join(
        TEMPLATE_DIR, "cell_template_neurodamus.jinja2"
    )
    hoc_template = os.path.abspath(hoc_template)
    emodels_in_repo = False
    continu = False

    with tools.cd(TEST_DATA_DIR):
        # create output directories and run function
        for path in [emodels_dir, hoc_dir]:
            tools.makedirs(path)
        arg_list = (original_emodel, emodel, emodel_dict, emodels_dir, opt_dir,
                    os.path.abspath(hoc_dir),
                    hoc_template, emodels_in_repo, continu)

        ret = prepare_emodel_dirs.prepare_emodel_dir(arg_list)

        # test side effects: creation of .hoc-file
        assert os.path.isdir(os.path.join(emodels_dir, emodel))
        hoc_path = os.path.join(hoc_dir, '{}.hoc'.format(emodel))
        assert os.path.isfile(hoc_path)

        # test returned dict
        expected_emodel_dir = os.path.join(emodels_dir, emodel)
        expected_ret = {emodel: expected_emodel_dir,
                        original_emodel: expected_emodel_dir}
        assert ret == expected_ret


@pytest.mark.unit
def test_prepare_emodel_dirs_single_process():
    """prepare_combos.prepare_emodel_dirs_single_process:
    test prepare_emodel_dirs based on test example 'simple1
    using a single process'.
    """
    # create test directory, where output of this test will be created
    test_dir = os.path.join(TMP_DIR, 'test_prepare_emodel_dirs')
    tools.makedirs(test_dir)

    # input parameters
    final_dict = {'emodel1': {'main_path': '.',
                              'seed': 2,
                              'rank': 0,
                              'notes': '',
                              'branch': 'emodel1',
                              'params': {'cm': 1.0},
                              'fitness': {'Step1.SpikeCount': 20.0},
                              'score': 104.72906197480131,
                              'morph_path': 'morphologies/morph1.asc'},
                  'emodel2': {'main_path': '.',
                              'seed': 2,
                              'rank': 0,
                              'notes': '',
                              'branch': 'emodel2',
                              'params': {'cm': 0.5},
                              'fitness': {'Step1.SpikeCount': 20.0},
                              'score': 104.72906197480131,
                              'morph_path': 'morphologies/morph2.asc'}}
    emodel_etype_map = {'emodel1': {'mm_recipe': 'emodel1',
                                    'etype': 'etype1',
                                    'layer': ['1', 'str1']},
                        'emodel2': {'mm_recipe': 'emodel2',
                                    'etype': 'etype2',
                                    'layer': ['1', '2']}}
    emodels_dir = os.path.join(test_dir, 'tmp/emodels/')
    opt_dir = os.path.join(TEST_DATA_DIR, 'data/emodels_dir/subdir/')
    emodels_hoc_dir = os.path.join(test_dir, './output/emodels_hoc/')
    hoc_template = os.path.join(
        TEMPLATE_DIR, "cell_template_neurodamus.jinja2"
    )
    hoc_template = os.path.abspath(hoc_template)
    emodels_in_repo = False
    continu = False

    # run function
    with tools.cd(TEST_DATA_DIR):
        ret = prepare_emodel_dirs.prepare_emodel_dirs(
            final_dict, emodel_etype_map, emodels_dir, opt_dir,
            emodels_hoc_dir, emodels_in_repo, hoc_template, continu,
            n_processes=1)

    # verify output
    expected_ret = {emodel: os.path.join(
        emodels_dir, emodel) for emodel in final_dict}
    assert ret == expected_ret
    assert os.path.isdir(emodels_dir)
    assert os.path.isdir(emodels_hoc_dir)


@pytest.mark.unit
def test_prepare_emodel_dirs_multi_process():
    """prepare_combos.prepare_emodel_dirs_multi_process:
    test prepare_emodel_dirs based on test example 'simple1'
    using multiprocessing.
    """
    # create test directory, where output of this test will be created
    test_dir = os.path.join(TMP_DIR, 'test_prepare_emodel_dirs')
    tools.makedirs(test_dir)

    # input parameters
    final_dict = {'emodel1': {'main_path': '.',
                              'seed': 2,
                              'rank': 0,
                              'notes': '',
                              'branch': 'emodel1',
                              'params': {'cm': 1.0},
                              'fitness': {'Step1.SpikeCount': 20.0},
                              'score': 104.72906197480131,
                              'morph_path': 'morphologies/morph1.asc'},
                  'emodel2': {'main_path': '.',
                              'seed': 2,
                              'rank': 0,
                              'notes': '',
                              'branch': 'emodel2',
                              'params': {'cm': 0.5},
                              'fitness': {'Step1.SpikeCount': 20.0},
                              'score': 104.72906197480131,
                              'morph_path': 'morphologies/morph2.asc'}}
    emodel_etype_map = {'emodel1': {'mm_recipe': 'emodel1',
                                    'etype': 'etype1',
                                    'layer': ['1', 'str1']},
                        'emodel2': {'mm_recipe': 'emodel2',
                                    'etype': 'etype2',
                                    'layer': ['1', '2']}}
    emodels_dir = os.path.join(test_dir, 'tmp/emodels/')
    opt_dir = os.path.join(TEST_DATA_DIR, 'data/emodels_dir/subdir/')
    emodels_hoc_dir = os.path.join(test_dir, './output/emodels_hoc/')
    hoc_template = os.path.join(
        TEMPLATE_DIR, "cell_template_neurodamus.jinja2"
    )
    hoc_template = os.path.abspath(hoc_template)
    emodels_in_repo = False
    continu = False

    # run function
    with tools.cd(TEST_DATA_DIR):
        ret = prepare_emodel_dirs.prepare_emodel_dirs(
            final_dict, emodel_etype_map, emodels_dir, opt_dir,
            emodels_hoc_dir, emodels_in_repo, hoc_template, continu,
            n_processes=None)

    # verify output
    expected_ret = {emodel: os.path.join(
        emodels_dir, emodel) for emodel in final_dict}
    assert ret == expected_ret
    assert os.path.isdir(emodels_dir)
    assert os.path.isdir(emodels_hoc_dir)



"""Tests for functionality in bluepymm/prepare_combos/create_mm_sqlite.py"""

"""
Copyright (c) 2018, EPFL/Blue Brain Project

 This file is part of BluePyMM <https://github.com/BlueBrain/BluePyMM>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


import pandas
import re
import os
import json

import pytest

from bluepymm.prepare_combos import create_mm_sqlite
from bluepymm import tools


BASE_DIR = os.path.abspath(os.path.dirname(__file__))
TEST_DIR = os.path.join(BASE_DIR, 'examples/simple1')


@pytest.mark.unit
def test_check_morphology_existence():
    """prepare_combos.create_mm_sqlite: tests check_morphology_existence"""
    morph_name = 'morph1.asc'
    morph_type = 'test'
    morph_dir = os.path.join(TEST_DIR, 'data/morphs', morph_name)
    ret = create_mm_sqlite.check_morphology_existence(morph_name, morph_type,
                                                      morph_dir)
    assert ret

    morph_name = 'does_not_exist.asc'
    morph_dir = os.path.join(TEST_DIR, 'data/morphs', morph_name)

    with pytest.raises(ValueError):
        create_mm_sqlite.check_morphology_existence(
            morph_name, morph_type, morph_dir)


@pytest.mark.unit
def test_create_exemplar_rows_skip_repaired_exemplar():
    """prepare_combos.create_mm_sqlite: test create_exemplar_rows
    based on test example 'simple1'.
    """
    emodel = 'emodel1'
    final_dict = {emodel: {
        'main_path': '.',
        'seed': 2,
        'rank': 0,
        'notes': '',
        'branch': 'emodel1',
        'params': {'cm': 1.0},
        'fitness': {'Step1.SpikeCount': 20.0},
        'score': 104.72906197480131,
        'morph_path': 'morphologies/morph1.asc'
    }}
    fullmtype_morph_map = {}  # not used in case of skip_repaired_exemplar
    emodel_etype_map = {emodel: {
        'mm_recipe': 'emodel1',
        'etype': 'etype1',
        'layer': ['1', 'str1']
    }}
    emodels = ['emodel1']
    emodel_dir = os.path.join(TEST_DIR, 'data/emodels_dir/subdir/')
    emodel_dirs = {emodel: emodel_dir}
    rep_morph_dir = os.path.join(TEST_DIR, 'data/morphs')
    skip_repaired_exemplar = True

    # construct expected output
    unrep_morph_dir = os.path.dirname(
        os.path.join(emodel_dirs[emodel], final_dict[emodel]['morph_path']))

    with tools.cd(TEST_DIR):
        ret = create_mm_sqlite.create_exemplar_rows(
            final_dict, fullmtype_morph_map, emodel_etype_map, emodels,
            emodel_dirs, rep_morph_dir, unrep_morph_dir,
            skip_repaired_exemplar)

    data = [(None, None, None, None, emodel_etype_map[emodel]['etype'],
             'morph1', '.asc', emodel, emodel, unrep_morph_dir, None,
             json.dumps(final_dict[emodel]['fitness']), None, True, True,
             False, False),
            (None, None, None, None, emodel_etype_map[emodel]['etype'],
             'morph1', '.asc', emodel, emodel, unrep_morph_dir, None,
             json.dumps(final_dict[emodel]['fitness']), None, True, True,
             False, True)]
    columns = [
        'layer',
        'fullmtype',
        'mtype',
        'msubtype',
        'etype',
        'morph_name',
        'morph_ext',
        'emodel',
        'original_emodel',
        'morph_dir',
        'scores',
        'opt_scores',
        'exception',
        'to_run',
        'is_exemplar',
        'is_repaired',
        'is_original']
    expected_ret = pandas.DataFrame(data, columns=columns)
    expected_ret.sort_index(axis=1, inplace=True)
    ret.sort_index(axis=1, inplace=True)

    pandas.testing.assert_frame_equal(ret, expected_ret)


@pytest.mark.unit
def test_remove_morph_regex_failures():
    """prepare_combos.create_mm_sqlite: test remove_morph_regex_failures"""
    data = pandas.DataFrame([('morph1', re.compile('morph1')),
                             ('morph2', re.compile('morph1')),
                             ('morph3', re.compile('.*')), ],
                            columns=['morph_name', 'morph_regex'])
    ret = create_mm_sqlite.remove_morph_regex_failures(data)

    expected_ret = pandas.DataFrame([('morph1'),
                                     ('morph3'), ],
                                    columns=['morph_name'])
    pandas.testing.assert_frame_equal(ret, expected_ret)


@pytest.mark.unit
def test_create_mm_sqlite():
    """prepare_combos.create_mm_sqlite: test create_mm_sqlite
    based on test example 'simple1'.
    """
    output_filename = 'scores.sqlite'
    recipe_filename = 'data/simple1_recipe.xml'
    morph_dir = 'data/morphs/'
    rep_morph_dir = 'data/rep_morphs/'
    emodel_dir = os.path.join(TEST_DIR, 'data/emodels_dir/subdir/')
    emodel_etype_map = tools.load_json(os.path.join(emodel_dir,
                                                    'emodel_etype_map.json'))
    final_dict = tools.load_json(os.path.join(emodel_dir, 'final.json'))
    emodel_dirs = {m: emodel_dir for m in ['emodel1', 'emodel2']}
    skip_repaired_exemplar = True

    with tools.cd(TEST_DIR):
        create_mm_sqlite.create_mm_sqlite(output_filename,
                                          recipe_filename,
                                          morph_dir,
                                          rep_morph_dir,
                                          rep_morph_dir,
                                          emodel_etype_map,
                                          final_dict,
                                          emodel_dirs,
                                          skip_repaired_exemplar)
        assert os.path.isfile(output_filename)
        # TODO: test database contents

        # clear output
        os.remove(output_filename)



"""Test bluepymm/prepare_combos"""

"""
Copyright (c) 2018, EPFL/Blue Brain Project

 This file is part of BluePyMM <https://github.com/BlueBrain/BluePyMM>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import os
import shutil
import csv

import pytest

import bluepymm


BASE_DIR = os.path.abspath(os.path.dirname(__file__))
TEST_DATA_DIR = os.path.join(BASE_DIR, 'examples/simple1')
TMP_DIR = os.path.join(BASE_DIR, 'tmp/test_legacy')


def teardown_module():
    """Remove the temporary files."""
    shutil.rmtree(TMP_DIR)


def _new_prepare_json(original_filename, test_dir):
    """Helper function to prepare new configuration file for prepare_combos."""
    config = bluepymm.tools.load_json(original_filename)
    config['tmp_dir'] = os.path.join(test_dir, 'tmp')
    config['output_dir'] = os.path.join(test_dir, 'output')
    config['scores_db'] = os.path.join(config['output_dir'], 'scores.sqlite')
    config['emodels_hoc_dir'] = os.path.join(config['output_dir'],
                                             'emodels_hoc')
    return bluepymm.tools.write_json(test_dir, original_filename, config)


def _prepare_config_jsons(prepare_config_template_filename,
                          hoc_config_template_filename):
    """Helper function to prepare configuration .json files."""
    # load json files
    prepare_config = bluepymm.tools.load_json(prepare_config_template_filename)
    hoc_config = bluepymm.tools.load_json(hoc_config_template_filename)

    # use TMP_DIR for output
    bluepymm.tools.makedirs(TMP_DIR)
    prepare_config['tmp_dir'] = os.path.abspath(os.path.join(TMP_DIR, 'tmp'))
    prepare_config['output_dir'] = os.path.abspath(os.path.join(TMP_DIR,
                                                                'output'))
    hoc_config['emodels_tmp_dir'] = os.path.join(prepare_config['tmp_dir'],
                                                 'emodels')
    hoc_config['hoc_output_dir'] = os.path.abspath(os.path.join(TMP_DIR,
                                                                'hoc_output'))

    # write out changes to TMP_DIR
    prepare_config_path = bluepymm.tools.write_json(
        TMP_DIR, prepare_config_template_filename, prepare_config)
    hoc_config_path = bluepymm.tools.write_json(
        TMP_DIR, hoc_config_template_filename, hoc_config)

    return prepare_config_path, hoc_config_path


@pytest.mark.unit
def test_get_parser():
    """bluepymm.legacy: test get_parser"""
    ret = bluepymm.legacy.create_hoc_files.get_parser()


@pytest.mark.unit
def test_add_full_paths():
    """bluepymm.legacy: test add_full_paths"""
    # input parameters
    test_dict = {'test': 'tmp/test_legacy/test_add_full_paths', 'int': 1}
    directory = BASE_DIR

    # prepare test data
    test_dir = os.path.join(TMP_DIR, 'test_add_full_paths')
    bluepymm.tools.makedirs(test_dir)

    # run function and verify output
    expected_ret = {'test': test_dir, 'int': 1}
    ret = bluepymm.legacy.create_hoc_files.add_full_paths(test_dict, directory)
    assert ret == expected_ret


@pytest.mark.unit
def test_load_combinations_dict():
    """bluepymm.legacy: test load_combinations_dict"""
    test_path = os.path.join(TEST_DATA_DIR,
                             'output_megate_expected/mecombo_emodel.tsv')

    emodel1_mtype1_1_morph1 = {'morph_name': 'morph1',
                               'layer': '1',
                               'fullmtype': 'mtype1',
                               'etype': 'etype1',
                               'emodel': 'emodel1',
                               'combo_name': 'emodel1_mtype1_1_morph1',
                               'threshold_current': '',
                               'holding_current': '',
                               }
    emodel1_mtype2_1_morph2 = {'morph_name': 'morph2',
                               'layer': '1',
                               'fullmtype': 'mtype2',
                               'etype': 'etype1',
                               'emodel': 'emodel1',
                               'combo_name': 'emodel1_mtype2_1_morph2',
                               'threshold_current': '',
                               'holding_current': '',
                               }
    emodel2_mtype1_1_morph1 = {'morph_name': 'morph1',
                               'layer': '1',
                               'fullmtype': 'mtype1',
                               'etype': 'etype2',
                               'emodel': 'emodel2',
                               'combo_name': 'emodel2_mtype1_1_morph1',
                               'threshold_current': '',
                               'holding_current': ''
                               }
    expected_ret = {'emodel1_mtype1_1_morph1': emodel1_mtype1_1_morph1,
                    'emodel1_mtype2_1_morph2': emodel1_mtype2_1_morph2,
                    'emodel2_mtype1_1_morph1': emodel2_mtype1_1_morph1,
                    }
    ret = bluepymm.legacy.create_hoc_files.load_combinations_dict(test_path)
    assert ret == expected_ret


@pytest.mark.unit
def test_run_create_and_write_hoc_file():
    """bluepymm.legacy: test run_create_and_write_hoc_file"""
    # create test directory, where output of this test will be created
    test_dir = os.path.join(TMP_DIR, 'test_run_create_and_write_hoc_file')
    bluepymm.tools.makedirs(test_dir)

    # input parameters
    prepare_conf_template = 'simple1_conf_prepare.json'

    emodel = 'emodel1'
    emodel_dir = os.path.join(test_dir, 'tmp', 'emodels', emodel)
    hoc_dir = os.path.join(test_dir, 'emodels_hoc')
    emodel_parameters = {'cm': 1.0}
    template = 'cell_template_neuron.jinja2'
    morph_path = 'morph.asc'
    model_name = 'test'

    with bluepymm.tools.cd(TEST_DATA_DIR):
        # prepare input files
        prepare_conf_path = _new_prepare_json(prepare_conf_template, test_dir)
        bluepymm.prepare_combos.main.prepare_combos(prepare_conf_path, False)

        # create output directory and run function
        bluepymm.tools.makedirs(hoc_dir)
        bluepymm.legacy.create_hoc_files.run_create_and_write_hoc_file(
            emodel, emodel_dir, hoc_dir, emodel_parameters, template,
            morph_path, model_name)

        # verify output directory. TODO: test hoc file contents
        expected_nb_files = 1
        list_of_files = os.listdir(hoc_dir)
        assert len(list_of_files) == expected_nb_files
        hoc_filename = '{}.hoc'.format(model_name)
        hoc_path = os.path.join(hoc_dir, hoc_filename)
        assert os.path.isfile(hoc_path)


@pytest.mark.unit
def test_create_hoc_files():
    """bluepymm.legacy: test create_hoc_files"""
    # create test directory, where output of this test will be created
    test_dir = os.path.join(TMP_DIR, 'test_create_hoc_files')
    bluepymm.tools.makedirs(test_dir)

    # input parameters
    prepare_conf_template = 'simple1_conf_prepare.json'

    model_name = 'combo1'
    combo_dict = {model_name: {'morph_name': 'morph1',
                               'layer': '1',
                               'fullmtype': 'mtype1',
                               'etype': 'etype1',
                               'emodel': 'emodel1',
                               'combo_name': 'combo1',
                               'threshold_current': '',
                               'holding_current': '',
                               },
                  }
    emodels_dir = os.path.join(test_dir, 'tmp', 'emodels')
    final_dict = {'emodel1': {'main_path': '.',
                              'seed': 2,
                              'rank': 0,
                              'notes': '',
                              'branch': 'emodel1',
                              'params': {'cm': 1.0},
                              'fitness': {'Step1.SpikeCount': 20.0},
                              'score': 104.72906197480131,
                              'morph_path': 'morphologies/morph1.asc'}
                  }
    template = 'cell_template.jinja2'
    hoc_dir = os.path.join(test_dir, 'emodels_hoc')

    with bluepymm.tools.cd(TEST_DATA_DIR):
        # prepare input files
        prepare_conf_path = _new_prepare_json(prepare_conf_template, test_dir)
        bluepymm.prepare_combos.main.prepare_combos(prepare_conf_path, False)

        # create output directory and run function
        bluepymm.tools.makedirs(hoc_dir)
        bluepymm.legacy.create_hoc_files.create_hoc_files(
            combo_dict, emodels_dir, final_dict, template, hoc_dir)

        # verify output directory. TODO: test hoc file contents
        expected_nb_files = 1
        list_of_files = os.listdir(hoc_dir)
        assert len(list_of_files) == expected_nb_files
        hoc_filename = '{}.hoc'.format(model_name)
        hoc_path = os.path.join(hoc_dir, hoc_filename)
        assert os.path.isfile(hoc_path)


def _verify_output(hoc_config_path):
    """Helper function to verify output"""
    hoc_config = bluepymm.tools.load_json(hoc_config_path)

    # verify .hoc-files existence - TODO: verify content
    assert os.path.isdir(hoc_config['hoc_output_dir'])
    with open(hoc_config['mecombo_emodel_filename']) as f:
        reader = csv.DictReader(f, delimiter='\t')
        for row in reader:
            hoc_path = os.path.join(hoc_config['hoc_output_dir'],
                                    '{}.hoc'.format(row['combo_name']))
            assert os.path.isfile(hoc_path)


def test_create_hoc_files_example_simple1():
    """bluepymm.legacy: test creation legacy .hoc files for example simple1"""
    prepare_config_template_filename = 'simple1_conf_prepare.json'
    hoc_config_template_filename = 'simple1_conf_hoc.json'
    with bluepymm.tools.cd(TEST_DATA_DIR):
        prepare_config_path, hoc_config_path = _prepare_config_jsons(
            prepare_config_template_filename, hoc_config_template_filename)

        # run combination preparation and create hoc files
        bluepymm.prepare_combos.main.prepare_combos(prepare_config_path, False)
        bluepymm.legacy.create_hoc_files.main([hoc_config_path])

        # verify output
        _verify_output(hoc_config_path)



"""Python Model Management"""

"""
Copyright (c) 2018, EPFL/Blue Brain Project

 This file is part of BluePyMM <https://github.com/BlueBrain/BluePyMM>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


# pylint: disable=C0325, W0223
# pylama: ignore=E402

import sys
import os
import json
import ipyparallel
import sqlite3
import traceback
import pandas

from bluepymm import tools


def run_emodel_morph_isolated(input_args):
    """Run e-model morphology combination in isolated environment.

    Args:
        input_args: tuple
        - uid: unique identifier of the e-model morphology combination
        - emodel: e-model name
        - emodel_dir: directory containing e-model files
        - emodel_params: dict that maps e-model parameters to their values
        - morph_path: path to morphology
        - apical_point_isec: integer value of the apical point isection
        - extra_values_error: boolean to raise an exception upon a missing key

    Returns:
        Dict with keys 'exception', 'extra_values', 'scores', 'uid'.
    """

    (
        uid,
        emodel,
        emodel_dir,
        emodel_params,
        morph_path,
        apical_point_isec,
        extra_values_error
    ) = input_args

    return_dict = {'uid': uid, 'exception': None}
    pool = tools.NestedPool(1, maxtasksperchild=1)

    try:
        return_dict['scores'], return_dict['extra_values'] = pool.apply(
            run_emodel_morph, (emodel,
                               emodel_dir,
                               emodel_params,
                               morph_path,
                               apical_point_isec,
                               extra_values_error
                               ))
    except Exception:
        return_dict['scores'] = None
        return_dict['extra_values'] = None
        return_dict['exception'] = "".join(traceback.format_exception(
                                           *sys.exc_info()))

    pool.terminate()
    pool.join()
    del pool

    return return_dict


def read_apical_point(morph_dir, morph_name):
    """Read apical point from apical point json file"""

    json_filename = os.path.join(morph_dir, 'apical_points_isec.json')

    with open(json_filename) as json_file:
        apic_points = json.load(json_file)

    # Get apic_point isec from dict, if not found return None
    if morph_name in apic_points:
        return int(apic_points[morph_name])
    else:
        return None


def run_emodel_morph(
        emodel,
        emodel_dir,
        emodel_params,
        morph_path,
        apical_point_isec,
        extra_values_error=True):
    """Run e-model morphology combination.

    Args:
        emodel: e-model name
        emodel_dir: directory containing e-model files
        emodel_params: dict that maps e-model parameters to their values
        morph_path: path to morphology
        apical_point_isec: integer value of the apical point isection
        extra_values_error: boolean to raise an exception upon a missing key

    Returns:
        tuple:
            - dict that maps features to scores
            - dict with extra values: 'holding_current' and 'threshold_current'
    """

    try:
        sys.stdout = open('/dev/null', 'w')
        print('Running e-model %s on morphology %s in %s' %
              (emodel, morph_path, emodel_dir))

        setup = tools.load_module('setup', emodel_dir)

        print("Changing path to %s" % emodel_dir)
        with tools.cd(emodel_dir):
            if hasattr(setup, 'multieval'):

                prefix = 'mm'
                altmorph = [[prefix, morph_path, apical_point_isec]]
                evaluator = setup.evaluator.create(etype='%s' % emodel,
                                                   altmorph=altmorph)

                evaluator = evaluator.evaluators[0]  # only one evaluator

                responses = evaluator.run_protocols(
                    evaluator.fitness_protocols.values(),
                    emodel_params)
                scores = evaluator.fitness_calculator.calculate_scores(
                    responses)

                extra_values = {}

                for response_key, extra_values_key in [
                        ('%s.bpo_holding_current' % prefix,
                         'holding_current'),
                        ('%s.bpo_threshold_current' % prefix,
                         'threshold_current')]:
                    if response_key in responses:
                        extra_values[extra_values_key] = responses[
                            response_key]
                    else:
                        if extra_values_error:
                            raise ValueError(
                                "Key %s not found in responses: %s" %
                                (response_key, str(responses)))
                        else:
                            extra_values[extra_values_key] = None
            else:
                evaluator = setup.evaluator.create(etype='%s' % emodel)
                evaluator.cell_model.morphology.morphology_path = morph_path

                responses = evaluator.run_protocols(
                    evaluator.fitness_protocols.values(),
                    emodel_params)
                scores = evaluator.fitness_calculator.calculate_scores(
                    responses)

                extra_values = {}
                extra_values['holding_current'] = \
                    responses.get('bpo_holding_current', None)
                extra_values['threshold_current'] = \
                    responses.get('bpo_threshold_current', None)

        return scores, extra_values
    except Exception:
        # Make sure exception and backtrace are thrown back to parent process
        raise Exception(
            "".join(traceback.format_exception(*sys.exc_info())))


def create_arg_list(scores_db_filename, emodel_dirs, final_dict,
                    extra_values_error=False, use_apical_points=True):
    """Create list of argument tuples to be used as an input for
    run_emodel_morph.

    Args:
        scores_db_filename: path to .sqlite database
        emodel_dirs: a dict mapping e-models to the directories with e-model
            input files
        final_dict: a dict mapping e-models to dicts with e-model parameters
        extra_values_error: boolean to raise an exception upon a missing key
        use_apical_points: boolean to use apical points or not

    Raises:
        ValueError, if one of the database entries contains has value None for
        the key 'emodel'.
    """

    arg_list = []
    with sqlite3.connect(scores_db_filename) as scores_db:
        scores_db.row_factory = sqlite3.Row

        one_row = scores_db.execute('SELECT * FROM scores LIMIT 1').fetchone()

        apical_points_isec = {}
        setup = tools.load_module('setup', emodel_dirs[one_row['emodel']])
        if hasattr(setup, 'multieval') and use_apical_points:
            apical_points_isec = tools.load_json(
                os.path.join(one_row['morph_dir'], "apical_points_isec.json")
            )

        scores_cursor = scores_db.execute('SELECT * FROM scores')
        for row in scores_cursor.fetchall():
            index = row['index']
            morph_name = row['morph_name']
            morph_ext = row['morph_ext']

            if morph_ext is None:
                morph_ext = '.asc'

            apical_point_isec = None
            if morph_name in apical_points_isec:
                apical_point_isec = int(apical_points_isec[morph_name])

            morph_filename = morph_name + morph_ext
            morph_path = os.path.abspath(os.path.join(row['morph_dir'],
                                                      morph_filename))
            if row['to_run'] == 1:
                emodel = row['emodel']
                original_emodel = row['original_emodel']
                if emodel is None:
                    raise ValueError(
                        "scores db row %s for morph %s, etype %s, mtype %s, "
                        "layer %s doesn't have an e-model assigned to it" %
                        (index, morph_name, row['etype'], row['mtype'],
                         row['layer']))
                args = (index, emodel,
                        os.path.abspath(emodel_dirs[emodel]),
                        final_dict[original_emodel]['params'],
                        morph_path, apical_point_isec, extra_values_error)
                arg_list.append(args)

    print('Found %d rows in score database to run' % len(arg_list))

    return arg_list


def save_scores(scores_db_filename, uid, scores, extra_values, exception,
                float_representation='.17g'):
    """Update a specific entry in a given database with scores and related
    parameters.

    Args:
        scores_db_filename: path to .sqlite database
        uid: unique identifier of database entry
        scores: scores dict to be added to entry as a json string
        extra_values: dict to be added to entry as a json string
        exception: description of exception that may have happened during score
                   calculation
        float_representation: use for json encoding. Default is '.17g'.

    Returns:
        ValueError if entry has already been updated.
    """
    json.encoder.FLOAT_REPR = lambda x: format(x, float_representation)

    with sqlite3.connect(scores_db_filename) as scores_db:
        # make sure we don't update a row that was already executed
        scores_cursor = scores_db.execute(
            'SELECT `index` FROM scores WHERE `index`=? AND to_run=?',
            (uid, False))
        if scores_cursor.fetchone() is None:
            # update row with calculated scores and related values
            scores_db.execute('UPDATE scores SET scores=?, extra_values=?, '
                              'exception=?, to_run=? WHERE `index`=?',
                              (json.dumps(scores), json.dumps(extra_values),
                               exception, False, uid))
        else:
            raise ValueError('save_scores: trying to update scores in a row '
                             'that was already executed: %d' % uid)


def expand_scores_to_score_values_table(scores_sqlite_filename):
    """Read scores from sqlite table, expand to dataframe, and store in new
    table 'score_values'. Each column of the new table corresponds to a
    single score.

    Args:
        scores_sqlite_filename: path to sqlite database with keys 'scores' and
                                'to_run'

    Raises:
        Exception, if the scores table contains at least one entry where the
        value of 'to_run' is True.
    """
    with sqlite3.connect(scores_sqlite_filename) as conn:
        scores = pandas.read_sql('SELECT * FROM scores', conn)
        tools.check_all_combos_have_run(scores, 'scores')

    score_values = scores['scores'].apply(
        lambda json_str: pandas.Series
        (json.loads(json_str)) if json_str else pandas.Series())

    with sqlite3.connect(scores_sqlite_filename) as conn:
        score_values.to_sql('score_values', conn, if_exists='replace',
                            index=False)


def calculate_scores(final_dict, emodel_dirs, scores_db_filename,
                     use_ipyp=False, ipyp_profile=None, timeout=10,
                     use_apical_points=True, n_processes=None):
    """Calculate scores of e-model morphology combinations and update the
    database accordingly.

    Args:
        scores_db_filename: path to .sqlite database with e-model morphology
            combinations
        final_dict: a dict mapping e-models to dicts with e-model parameters
        emodel_dirs: a dict mapping e-models to the directories with e-model
            input files
        use_ipyp: bool indicating whether ipyparallel is used. Default is
            False.
        ipyp_profile: path to ipyparallel profile. Default is None.
        use_apical_points: boolean to use apical points or not
        n_processes: the integer number of processes. If `None`,
        all processes are going to be used.
    """

    print('Creating argument list for parallelisation')
    arg_list = create_arg_list(scores_db_filename,
                               emodel_dirs,
                               final_dict,
                               use_apical_points=use_apical_points)

    print('Parallelising score evaluation of %d me-combos' % len(arg_list))
    if use_ipyp:
        # use ipyparallel
        client = ipyparallel.Client(profile=ipyp_profile, timeout=timeout)
        lview = client.load_balanced_view(targets=n_processes)
        results = lview.imap(run_emodel_morph_isolated,
                             arg_list, ordered=False)
    else:
        # use multiprocessing
        pool = tools.NestedPool(processes=n_processes)
        results = pool.imap_unordered(run_emodel_morph_isolated, arg_list)

    # every time a result comes in, save the score in the database
    for uids_received, result in enumerate(results, start=1):
        uid = result['uid']
        scores = result['scores']
        extra_values = result['extra_values']
        exception = result['exception']
        save_scores(scores_db_filename, uid, scores, extra_values, exception)

        print('Saved scores for uid %s (%d out of %d) %s' %
              (uid, uids_received, len(arg_list),
               'with exception' if exception else ''))
        sys.stdout.flush()

    if not use_ipyp:
        pool.terminate()
        pool.join()

    print('Converting score json strings to scores values ...')
    expand_scores_to_score_values_table(scores_db_filename)



"""Test process_megate_config"""

"""
Copyright (c) 2018, EPFL/Blue Brain Project

 This file is part of BluePyMM <https://github.com/BlueBrain/BluePyMM>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


import re

import pytest

from bluepymm.select_combos import process_megate_config as proc_config


@pytest.mark.unit
def test_join_regex():
    """select_combos.process_megate_config: test join_regex"""
    test_list = ['one', '.*', 'three']
    joined_list = '(one$)|(.*$)|(three$)'
    ret = proc_config.join_regex(test_list)
    assert ret == re.compile(joined_list)


def _test_read_to_skip_features(skip_features, conf_dict):
    """Test read_to_skip_features helper function"""
    r_patterns, r_features = proc_config.read_to_skip_features(conf_dict)
    assert skip_features == r_features
    exp_patterns = [re.compile(f) for f in skip_features]
    assert exp_patterns == r_patterns


@pytest.mark.unit
def test_read_to_skip_features():
    """select_combos.process_megate_config: test read_to_skip_features"""

    skip_features = []
    conf_dict = {'to_skip_features': skip_features}
    _test_read_to_skip_features(skip_features, conf_dict)

    skip_features = []
    conf_dict = {}
    _test_read_to_skip_features(skip_features, conf_dict)

    skip_features = ['test']
    conf_dict = {'to_skip_features': skip_features}
    _test_read_to_skip_features(skip_features, conf_dict)

    skip_features = ['.*']
    conf_dict = {'to_skip_features': skip_features}
    _test_read_to_skip_features(skip_features, conf_dict)


'''
# Disabling this test for now because it is unstable (give stochastic results)
@pytest.mark.unit
def test_read_megate_thresholds():
    """select_combos.process_megate_config: test read_megate_thresholds"""

    # all keys present
    test_dict = {'megate_thresholds': [
        {'emodel': ['test1'], 'fullmtype': ['test2'], 'etype': ['test3'],
         'features': ['.*'], 'megate_threshold': 5}]}
    ret_patterns, ret_thresholds = proc_config.read_megate_thresholds(
        test_dict)
    expected_patterns = [
        {'megate_feature_threshold':
         {'megate_threshold': 5, 'features': proc_config.join_regex(['.*'])},
         'emodel': proc_config.join_regex(['test1']), 'fullmtype':
         proc_config.join_regex(['test2']), 'etype': proc_config.join_regex(
             ['test3'])}]
    nt.assert_list_equal(ret_thresholds, test_dict['megate_thresholds'])
    nt.assert_equal(len(ret_patterns), len(expected_patterns))
    nt.assert_dict_equal(ret_patterns[0], expected_patterns[0])

    # key 'fullmtype' not present
    test_dict = {'megate_thresholds': [
        {'emodel': ['test1'], 'etype': ['test3'], 'features': ['.*'],
         'megate_threshold': 5}]}
    ret_patterns, ret_thresholds = proc_config.read_megate_thresholds(
        test_dict)
    expected_patterns = [
        {'megate_feature_threshold':
         {'megate_threshold': 5, 'features': proc_config.join_regex(['.*'])},
         'emodel': proc_config.join_regex(['test1']), 'fullmtype':
         re.compile('.*'), 'etype': proc_config.join_regex(['test3'])}]
    nt.assert_list_equal(ret_thresholds, test_dict['megate_thresholds'])
    nt.assert_equal(len(ret_patterns), len(expected_patterns))
    nt.assert_dict_equal(ret_patterns[0], expected_patterns[0])
'''



"""BluePyMM megate output."""

"""
Copyright (c) 2018, EPFL/Blue Brain Project

 This file is part of BluePyMM <https://github.com/BlueBrain/BluePyMM>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


# pylint: disable=R0914, C0325, W0640


import os
from bluepymm import tools
from . import table_processing


def _write_extneurondbdat(extneurondb, filename):
    """Write extneurondb.dat"""
    pure_extneuron_db = extneurondb.copy()

    # Select the correct columns
    column_order = ['morph_name', 'layer', 'fullmtype', 'etype', 'combo_name']
    # pure_extneuron_db = pure_extneuron_db[column_order]
    pure_extneuron_db.to_csv(
        filename,
        sep=' ',
        columns=column_order,
        index=False,
        header=False)


def save_megate_results(extneurondb, output_dir,
                        extneurondb_filename='extneurondb.dat',
                        mecombo_emodel_filename='mecombo_emodel.tsv',
                        sort_key=None,
                        make_names_neuron_compliant=False,
                        extra_value_errors=True):
    """Write results of megating to two files.

    Args:
        extneurondb: pandas.DataFrame with result of me-gating
        output_dir: path to output directory
        extneurondb_filename: filename of extended neuron database. The columns
                              of this file are ordered as 'morph_name',
                              'layer', 'fullmtype', 'etype', 'combo_name'.
                              Values are separated by a space. Default filename
                              is 'extneurondb.dat'.
        mecombo_emodel_filename: filename of 'mecombo_emodel' file. Values are
                                 separated with a tab. Default filename is
                                 'mecombo_emodel.tsv'.
        sort_key: key to sort database in ascending order before writing out to
                  file. Default is None.
        make_names_neuron_compliant: boolean indicating whether the combo name
                                     should be made NEURON-compliant. Default
                                     is False. If set to True, a log file with
                                     the conversion info is written out to
                                     <output_dir>/log_neuron_compliance.csv
    """
    tools.makedirs(output_dir)

    if make_names_neuron_compliant:
        log_filename = 'log_neuron_compliance.csv'
        log_path = os.path.join(output_dir, log_filename)
        table_processing.process_combo_name(extneurondb, log_path)

    if sort_key is not None:
        extneurondb = extneurondb.sort_values(sort_key).reset_index(drop=True)

    extneurondb_path = os.path.join(output_dir, extneurondb_filename)
    _write_extneurondbdat(extneurondb, extneurondb_path)
    print(
        'Wrote extneurondb.dat to {}'.format(
            os.path.abspath(extneurondb_path)))

    mecombo_emodel_path = os.path.join(output_dir, mecombo_emodel_filename)

    if extra_value_errors:
        for extra_values_key in ['holding_current', 'threshold_current']:
            null_rows = extneurondb[extra_values_key].isnull()
            if null_rows.sum() > 0:
                # TODO reenable this for release !
                # raise ValueError(
                #    "There are rows with None for "
                #    "holding current: %s" % str(
                #        extneurondb[null_rows]))
                print("WARNING ! There are rows with None for "
                      "holding current: %s" % str(extneurondb[null_rows]))

    extneurondb.to_csv(
        mecombo_emodel_path,
        columns=[
            'morph_name',
            'layer',
            'fullmtype',
            'etype',
            'emodel',
            'combo_name',
            'threshold_current',
            'holding_current'],
        sep='\t',
        index=False)
    print(
        'Wrote mecombo_emodel tsv to {}'.format(
            os.path.abspath(mecombo_emodel_path)))

    return extneurondb_path, mecombo_emodel_path


def write_mecomboreleasejson(
        output_dir,
        emodels_hoc_path,
        extneurondb_path,
        mecombo_emodel_path):
    """Write json file contain info about release"""

    output_paths = {}
    output_paths['emodels_hoc'] = os.path.abspath(emodels_hoc_path)
    output_paths['extneurondb.dat'] = os.path.abspath(extneurondb_path)
    output_paths['mecombo_emodel.tsv'] = os.path.abspath(mecombo_emodel_path)
    release = {'version': '1.0', 'output_paths': output_paths}
    tools.write_json(
        output_dir,
        'mecombo_release.json',
        release)

    print(
        'Wrote mecombo_release json to %s' %
        os.path.abspath(os.path.join(
            output_dir,
            'mecombo_release.json')))



"""Tests for functionality in bluepymm/select_combos/sqlite_io.py"""

"""
Copyright (c) 2018, EPFL/Blue Brain Project

 This file is part of BluePyMM <https://github.com/BlueBrain/BluePyMM>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


import sqlite3
import os
import pandas

import pytest

from bluepymm.select_combos import sqlite_io
from bluepymm import tools


BASE_DIR = os.path.abspath(os.path.dirname(__file__))
TMP_DIR = os.path.join(BASE_DIR, 'tmp/test_sqlite_io')


def _create_database(test_dir, filename, scores, score_values):
    """Helper function to create test database."""
    path = os.path.join(test_dir, filename)
    with sqlite3.connect(path) as conn:
        scores.to_sql('scores', conn, if_exists='replace', index=False)
        score_values.to_sql('score_values', conn, if_exists='replace')
    return path


@pytest.mark.unit
def test_read_and_process_sqlite_score_tables():
    """select_combos.sqlite_io: test read_and_process_sqlite_score_tables"""
    # create database
    scores_row = {'test': 1}
    scores = pandas.DataFrame(scores_row, index=[0])
    score_values_row = {'value': 2}
    score_values = pandas.DataFrame(score_values_row, index=[0])
    test_dir = os.path.join(
        TMP_DIR, 'test_read_and_process_sqlite_score_tables')
    tools.makedirs(test_dir)
    filename = 'test_db.sql'
    path = _create_database(test_dir, filename, scores, score_values)

    # read database
    ret_scs, ret_sc_vals = sqlite_io.read_and_process_sqlite_score_tables(path)

    # verify output
    assert 'index' not in ret_sc_vals.columns.values

    pandas.testing.assert_frame_equal(ret_scs, scores)
    pandas.testing.assert_frame_equal(ret_sc_vals, score_values)


@pytest.mark.unit
def test_read_and_process_sqlite_score_tables_error():
    """select_combos.sqlite_io: test read_and_process_sqlite_score_tables excep
    """
    # create database, table 'scores' has one entry, table 'score_values' two
    scores_row = {'test': [1, 3]}
    scores = pandas.DataFrame(scores_row)
    score_values_row = {'value': 2}
    score_values = pandas.DataFrame(score_values_row, index=[0])
    test_dir = os.path.join(
        TMP_DIR, 'test_read_and_process_sqlite_score_tables_error')
    tools.makedirs(test_dir)
    filename = 'test_db_error.sql'
    path = _create_database(test_dir, filename, scores, score_values)

    # read database, number of rows incompatible -> exception
    with pytest.raises(Exception):
        sqlite_io.read_and_process_sqlite_score_tables(path)



"""Tests for calculate_scores"""

from __future__ import print_function

"""
Copyright (c) 2018, EPFL/Blue Brain Project

 This file is part of BluePyMM <https://github.com/BlueBrain/BluePyMM>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


import os
import pandas
import sqlite3
import ipyparallel as ipp
import json
import subprocess
import time

import pytest

import bluepymm.run_combos as run_combos
from bluepymm import tools


BASE_DIR = os.path.abspath(os.path.dirname(__file__))
TEST_DIR = os.path.join(BASE_DIR, 'examples/simple1')
TMP_DIR = os.path.join(BASE_DIR, 'tmp/')


@pytest.mark.unit
def test_run_emodel_morph_isolated():
    """run_combos.calculate_scores: test run_emodel_morph_isolated."""
    uid = 0
    emodel = 'emodel1'
    emodel_dir = os.path.join(TEST_DIR, 'data/emodels_dir/subdir/')
    emodel_params = {'cm': 1.0}
    morph_name = 'morph1'
    morph_dir = os.path.join(TEST_DIR, 'data/morphs')
    morph_path = os.path.join(morph_dir, '%s.asc' % morph_name)

    input_args = (
        uid,
        emodel,
        emodel_dir,
        emodel_params,
        morph_path,
        None,
        False)
    ret = run_combos.calculate_scores.run_emodel_morph_isolated(input_args)

    expected_ret = {'exception': None,
                    'extra_values': {'holding_current': None,
                                     'threshold_current': None},
                    'scores': {'Step1.SpikeCount': 20.0},
                    'uid': 0}
    assert ret == expected_ret


@pytest.mark.unit
def test_run_emodel_morph_isolated_exception():
    """run_combos.calculate_scores: test run_emodel_morph_isolated exception.
    """
    # input parameters
    uid = 0
    emodel = 'Key mm.bpo_holding_current not found in responses'
    emodel_dir = os.path.join(TEST_DIR, 'data/emodels_dir/subdir/')
    emodel_params = {'cm': 1.0}
    morph_name = 'morph1'
    morph_dir = os.path.join(TEST_DIR, 'data/morphs')
    morph_path = os.path.join(morph_dir, '%s.asc' % morph_name)

    # function call
    input_args = (
        uid,
        emodel,
        emodel_dir,
        emodel_params,
        morph_path,
        None,
        True)
    ret = run_combos.calculate_scores.run_emodel_morph_isolated(input_args)

    # verify output: exception thrown because of non-existing e-model
    expected_ret = {'exception': 'this_is_a_placeholder',
                    'extra_values': None,
                    'scores': None,
                    'uid': 0}
    assert ret.keys() == expected_ret.keys()
    for k in ['extra_values', 'scores', 'uid']:
        assert ret[k] == expected_ret[k]
    assert emodel in ret['exception']


@pytest.mark.unit
def test_run_emodel_morph():
    """run_combos.calculate_scores: test run_emodel_morph."""
    emodel = 'emodel1'
    emodel_dir = os.path.join(TEST_DIR, 'data/emodels_dir/subdir/')
    emodel_params = {'cm': 1.0}

    morph_name = 'morph1'
    morph_dir = os.path.join(TEST_DIR, 'data/morphs')
    morph_path = os.path.join(morph_dir, '%s.asc' % morph_name)

    ret = run_combos.calculate_scores.run_emodel_morph(
        emodel,
        emodel_dir,
        emodel_params,
        morph_path,
        None,
        False)

    expected_scores = {'Step1.SpikeCount': 20.0}
    expected_extra_values = {'holding_current': None,
                             'threshold_current': None}
    assert ret[0] == expected_scores
    assert ret[1] == expected_extra_values


@pytest.mark.unit
def test_run_emodel_morph_exception():
    """run_combos.calculate_scores: test run_emodel_morph exception."""
    emodel = 'emodel_doesnt_exist'
    emodel_dir = os.path.join(TEST_DIR, 'data/emodels_dir/subdir/')
    emodel_params = {'cm': 1.0}

    morph_name = 'morph1'
    morph_dir = os.path.join(TEST_DIR, 'data/morphs')
    morph_path = os.path.join(morph_dir, '%s.asc' % morph_name)

    with pytest.raises(Exception):
        run_combos.calculate_scores.run_emodel_morph(
            emodel,
            emodel_dir,
            emodel_params,
            morph_path)


def _write_test_scores_database(row, testsqlite_filename):
    """Helper function to create test scores database."""
    df = pandas.DataFrame(row, index=[0])
    with sqlite3.connect(testsqlite_filename) as conn:
        df.to_sql('scores', conn, if_exists='replace')


@pytest.mark.unit
def test_create_arg_list():
    """run_combos.calculate_scores: test create_arg_list."""
    # write database
    testsqlite_filename = os.path.join(TMP_DIR, 'test1.sqlite')
    index = 0
    morph_name = 'morph'
    morph_dir = os.path.join(TEST_DIR, 'data/morphs')
    mtype = 'mtype'
    etype = 'etype'
    layer = 'layer'
    emodel = 'emodel'
    emodel_dir = os.path.join(TEST_DIR, 'data/emodels_dir/subdir/')
    row = {'index': index,
           'morph_name': morph_name,
           'morph_ext': None,
           'morph_dir': morph_dir,
           'mtype': mtype,
           'etype': etype,
           'layer': layer,
           'emodel': emodel,
           'original_emodel': emodel,
           'to_run': 1}
    _write_test_scores_database(row, testsqlite_filename)

    # extra input parameters
    emodel_dirs = {emodel: emodel_dir}
    params = 'test'
    final_dict = {emodel: {'params': params}}
    extra_values_error = False
    # from nose.tools import set_trace; set_trace()
    ret = run_combos.calculate_scores.create_arg_list(
        testsqlite_filename,
        emodel_dirs,
        final_dict,
        extra_values_error)

    # verify output
    morph_path = os.path.join(morph_dir, '{}.asc'.format(morph_name))
    expected_ret = [(index,
                     emodel,
                     os.path.abspath(emodel_dirs[emodel]),
                     params,
                     os.path.abspath(morph_path),
                     0, extra_values_error)]
    assert ret == expected_ret


@pytest.mark.unit
def test_create_arg_list_exception():
    """run_combos.calculate_scores: test create_arg_list for ValueError."""
    # write database
    testsqlite_filename = os.path.join(TMP_DIR, 'test2.sqlite')
    index = 0
    morph_name = 'morph'
    morph_dir = os.path.join(TEST_DIR, 'data/morphs')
    mtype = 'mtype'
    etype = 'etype'
    layer = 'layer'
    emodel = None
    emodel_dir = os.path.join(TEST_DIR, 'data/emodels_dir/subdir/')
    row = {'index': index,
           'morph_name': morph_name,
           'morph_ext': None,
           'morph_dir': morph_dir,
           'mtype': mtype,
           'etype': etype,
           'layer': layer,
           'emodel': emodel,
           'original_emodel': emodel,
           'to_run': 1}
    _write_test_scores_database(row, testsqlite_filename)

    # extra input parameters
    emodel_dirs = {emodel: emodel_dir}
    params = 'test'
    final_dict = {emodel: {'params': params}}

    # emodel is None -> raises ValueError
    with pytest.raises(ValueError):
        run_combos.calculate_scores.create_arg_list(
            testsqlite_filename,
            emodel_dirs,
            final_dict)


def _dict_factory(cursor, row):
    """Helper function to create dictionaries from database rows."""
    return {col[0]: row[idx] for idx, col in enumerate(cursor.description)}


@pytest.mark.unit
def test_save_scores():
    """run_combos.calculate_scores: test save_scores"""
    # create test database with single entry 'row'
    testsqlite_filename = os.path.join(TMP_DIR, 'test3.sqlite')
    row = {'scores': None,
           'extra_values': None,
           'exception': None,
           'to_run': True}
    _write_test_scores_database(row, testsqlite_filename)

    # process database
    uid = 0
    scores = {'score': 1}
    extra_values = {'extra': 2}
    exception = 'exception'
    run_combos.calculate_scores.save_scores(
        testsqlite_filename,
        uid,
        scores,
        extra_values,
        exception)

    # verify database
    expected_db_row = {'index': uid,
                       'scores': json.dumps(scores),
                       'extra_values': json.dumps(extra_values),
                       'exception': exception,
                       'to_run': 0}  # False
    with sqlite3.connect(testsqlite_filename) as scores_db:
        scores_db.row_factory = _dict_factory
        scores_cursor = scores_db.execute('SELECT * FROM scores')
        db_row = scores_cursor.fetchall()[0]
        assert db_row == expected_db_row

    # value already updated -> error
    with pytest.raises(ValueError):
        run_combos.calculate_scores.save_scores(
            testsqlite_filename, uid, scores, extra_values, exception
        )


@pytest.mark.unit
def test_expand_scores_to_score_values_table():
    """run_combos.calculate_scores: test expand_scores_to_score_values_table"""
    # create database
    db_path = os.path.join(TMP_DIR, 'test_expand_scores.sqlite')
    score_key = 'Step1.SpikeCount'
    score_value = 20.0
    scores = '{"%s": %s}' % (score_key, score_value)
    row = {'scores': scores, 'to_run': False}
    _write_test_scores_database(row, db_path)

    # process database
    run_combos.calculate_scores.expand_scores_to_score_values_table(db_path)

    # verify database
    expected_df = pandas.DataFrame(data=json.loads(scores), index=[0])
    with sqlite3.connect(db_path) as conn:
        score_values = pandas.read_sql('SELECT * FROM score_values', conn)
    pandas.testing.assert_frame_equal(score_values, expected_df)


@pytest.mark.unit
def test_expand_scores_to_score_values_table_error():
    """run_combos.calculate_scores: test expand_scores_to_score_values_table 2
    """
    # create database
    db_path = os.path.join(TMP_DIR, 'test_expand_scores_error.sqlite')
    score_key = 'Step1.SpikeCount'
    score_value = 20.0
    scores = '{"%s": %s}' % (score_key, score_value)
    row = {'scores': scores, 'to_run': True}
    _write_test_scores_database(row, db_path)

    # process database
    with pytest.raises(Exception):
        run_combos.calculate_scores.expand_scores_to_score_values_table(
            db_path)


@pytest.fixture()
def ipp_cluster_fixture():
    """Starts and terminates the ipcluster engine."""
    ip_proc = subprocess.Popen(["ipcluster", "start", "-n=2"])
    # ensure that ipcluster has enough time to start
    time.sleep(15)
    yield
    ip_proc.terminate()


@pytest.mark.unit
def test_calculate_scores(ipp_cluster_fixture):
    """run_combos.calculate_scores: test calculate_scores"""
    # write database
    test_db_filename = os.path.join(TMP_DIR, 'test4.sqlite')
    morph_name = 'morph1'
    morph_dir = os.path.join(TEST_DIR, 'data/morphs')
    mtype = 'mtype1'
    etype = 'etype1'
    layer = 1
    emodel = 'emodel1'
    exception = None
    row = {'morph_name': morph_name,
           'morph_ext': None,
           'morph_dir': morph_dir,
           'mtype': mtype,
           'etype': etype,
           'layer': layer,
           'emodel': emodel,
           'original_emodel': emodel,
           'to_run': 1,
           'scores': None,
           'extra_values': None,
           'exception': exception}
    _write_test_scores_database(row, test_db_filename)

    # calculate scores
    emodel_dir = os.path.join(TEST_DIR, 'data/emodels_dir/subdir/')
    emodel_dirs = {emodel: emodel_dir}
    final_dict_path = os.path.join(emodel_dir, 'final.json')
    final_dict = tools.load_json(final_dict_path)

    for use_ipyp in [False, True]:
        with tools.cd(TEST_DIR):
            run_combos.calculate_scores.calculate_scores(
                final_dict,
                emodel_dirs,
                test_db_filename,
                use_ipyp=use_ipyp,
                n_processes=1
            )

        # verify database
        scores = {'Step1.SpikeCount': 20.0}
        extra_values = {'holding_current': None, 'threshold_current': None}
        expected_db_row = {'index': 0,
                           'morph_name': morph_name,
                           'morph_ext': None,
                           'morph_dir': morph_dir,
                           'mtype': mtype,
                           'etype': etype,
                           'layer': layer,
                           'emodel': emodel,
                           'original_emodel': emodel,
                           'to_run': 0,
                           'scores': json.dumps(scores),
                           'extra_values': json.dumps(extra_values),
                           'exception': exception}
        with sqlite3.connect(test_db_filename) as scores_db:
            scores_db.row_factory = _dict_factory
            scores_cursor = scores_db.execute('SELECT * FROM scores')
            db_row = scores_cursor.fetchall()[0]
            assert db_row == expected_db_row


@pytest.mark.unit
def test_read_apical_point():
    """run_combos.calculate_scores: test read_apical_point."""
    morph_name = 'morph'
    morph_dir = os.path.join(TEST_DIR, 'data/morphs')

    apical_point_isec = run_combos.calculate_scores.read_apical_point(
        morph_dir, morph_name)

    assert apical_point_isec == 0



"""Tests for select_combos/megate_output"""

from __future__ import print_function

"""
Copyright (c) 2018, EPFL/Blue Brain Project

 This file is part of BluePyMM <https://github.com/BlueBrain/BluePyMM>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import pandas
import filecmp
import os

import pytest

import bluepymm.select_combos as select_combos
from bluepymm import tools


BASE_DIR = os.path.abspath(os.path.dirname(__file__))
TEST_DATA_DIR = os.path.join(BASE_DIR, 'examples/simple1')
TMP_DIR = os.path.join(BASE_DIR, 'tmp/megate_output')


def _test_save_megate_results(data, sort_key, test_dir, compliant):
    # input parameters
    columns = ['morph_name', 'layer', 'fullmtype', 'etype', 'emodel',
               'combo_name', 'threshold_current', 'holding_current']
    df = pandas.DataFrame(data, columns=columns)

    # save_megate_results
    select_combos.megate_output.save_megate_results(
        df, test_dir, sort_key=sort_key, make_names_neuron_compliant=compliant)

    # verify output files
    benchmark_dir = os.path.join(TEST_DATA_DIR, 'output_megate_expected')
    files = ['extneurondb.dat', 'mecombo_emodel.tsv']
    matches = filecmp.cmpfiles(benchmark_dir, test_dir, files)
    if len(matches[0]) != len(files):
        print('Mismatch in files: {}'.format(matches[1]))
    assert len(matches[0]) == len(files)

    if compliant:
        logfile_path = os.path.join(test_dir, 'log_neuron_compliance.csv')
        assert os.path.isfile(logfile_path)


@pytest.mark.unit
def test_save_megate_results_no_sort():
    """bluepymm.select_combos: test save_megate_results."""
    data = [('morph1', 1, 'mtype1', 'etype1', 'emodel1',
             'emodel1_mtype1_1_morph1', '', ''),
            ('morph2', 1, 'mtype2', 'etype1', 'emodel1',
             'emodel1_mtype2_1_morph2', '', ''),
            ('morph1', 1, 'mtype1', 'etype2', 'emodel2',
             'emodel2_mtype1_1_morph1', '', '')]
    test_dir = os.path.join(TMP_DIR, 'test_save_megate_results_no_sort')
    tools.makedirs(test_dir)
    _test_save_megate_results(data, None, test_dir, False)


@pytest.mark.unit
def test_save_megate_results_sort():
    """bluepymm.select_combos: test save_megate_results sorted."""
    data = [('morph1', 1, 'mtype1', 'etype1', 'emodel1',
             'emodel1_mtype1_1_morph1', '', ''),
            ('morph1', 1, 'mtype1', 'etype2', 'emodel2',
             'emodel2_mtype1_1_morph1', '', ''),
            ('morph2', 1, 'mtype2', 'etype1', 'emodel1',
             'emodel1_mtype2_1_morph2', '', '')]
    test_dir = os.path.join(TMP_DIR, 'test_save_megate_results_sort')
    tools.makedirs(test_dir)
    _test_save_megate_results(data, 'combo_name', test_dir, False)


@pytest.mark.unit
def test_save_megate_results_compliant():
    """bluepymm.select_combos: test save_megate_results neuron compliant."""
    data = [('morph1', 1, 'mtype1', 'etype1', 'emodel1',
             'emodel1_mtype1_1_morph1', '', ''),
            ('morph2', 1, 'mtype2', 'etype1', 'emodel1',
             'emodel1_mtype2_1_morph2', '', ''),
            ('morph1', 1, 'mtype1', 'etype2', 'emodel2',
             'emodel2_mtype1_1_morph1', '', '')]
    test_dir = os.path.join(TMP_DIR, 'test_save_megate_results_compliant')
    tools.makedirs(test_dir)
    _test_save_megate_results(data, None, test_dir, True)



""" Create database of possible me-combinations."""

from __future__ import print_function

"""
Copyright (c) 2018, EPFL/Blue Brain Project

 This file is part of BluePyMM <https://github.com/BlueBrain/BluePyMM>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


import os

from bluepymm import tools
from . import prepare_emodel_dirs as prepare_dirs
from . import create_mm_sqlite


def prepare_emodels(conf_dict, continu, scores_db_path, n_processes):
    """Prepare emodels"""

    tmp_dir = conf_dict['tmp_dir']
    emodels_dir = os.path.abspath(os.path.join(tmp_dir, 'emodels'))

    # Convert e-models input to BluePyMM file structure
    emodels_in_repo = prepare_dirs.check_emodels_in_repo(conf_dict)
    tmp_emodels_dir = prepare_dirs.convert_emodel_input(emodels_in_repo,
                                                        conf_dict,
                                                        continu)

    # Get information from emodels repo
    print('Getting final emodels dict')
    final_dict, emodel_etype_map, opt_dir = prepare_dirs.get_emodel_dicts(
        tmp_emodels_dir, conf_dict['final_json_path'],
        conf_dict['emodel_etype_map_path'])

    if "template" in conf_dict.keys():
        hoc_template = os.path.abspath(conf_dict["template"])
    else:
        base_dir = os.path.abspath(os.path.dirname(__file__))
        template_dir = os.path.join(base_dir, '../templates')
        hoc_template = os.path.join(
            template_dir, 'cell_template_neurodamus.jinja2')
        hoc_template = os.path.abspath(hoc_template)
    print('Preparing emodels in %s' % emodels_dir)
    emodels_hoc_dir = os.path.abspath(conf_dict['emodels_hoc_dir'])
    # Clone the emodels repo and prepare the dirs for all the emodels
    emodel_dirs = prepare_dirs.prepare_emodel_dirs(
        final_dict, emodel_etype_map, emodels_dir, opt_dir, emodels_hoc_dir,
        emodels_in_repo, hoc_template, continu=continu,
        n_processes=n_processes)

    if not continu:
        print('Creating sqlite db at %s' % scores_db_path)
        skip_repaired_exemplar = conf_dict.get('skip_repaired_exemplar', False)
        morph_dir = conf_dict['morph_path']
        rep_morph_dir = conf_dict['rep_morph_path']
        unrep_morph_dir = conf_dict.get('unrep_morph_path', None)
        print('Using repaired exemplar morph path: %s' % rep_morph_dir)
        print('Using unrepaired exemplar morph path: %s' % unrep_morph_dir)

        if 'circuitmvd3_path' in conf_dict:
            if 'recipe_path' in conf_dict:
                raise ValueError('Impossible to specify both recipe_path '
                                 'and circuitmvd3_path in config file')
            circuitmvd3_path = conf_dict['circuitmvd3_path']

            create_mm_sqlite.create_mm_sqlite_circuitmvd3(
                scores_db_path,
                circuitmvd3_path,
                morph_dir,
                rep_morph_dir,
                unrep_morph_dir,
                emodel_etype_map,
                final_dict,
                emodel_dirs,
                skip_repaired_exemplar=skip_repaired_exemplar)
        else:
            recipe_filename = conf_dict['recipe_path']

            # Create a sqlite3 db with all the combos
            create_mm_sqlite.create_mm_sqlite(
                scores_db_path,
                recipe_filename,
                morph_dir,
                rep_morph_dir,
                unrep_morph_dir,
                emodel_etype_map,
                final_dict,
                emodel_dirs,
                skip_repaired_exemplar=skip_repaired_exemplar)

    return final_dict, emodel_dirs


def prepare_combos(conf_filename, continu, n_processes=None):
    """Prepare combos"""

    print('Reading configuration at %s' % conf_filename)
    conf_dict = tools.load_json(conf_filename)
    scores_db_path = os.path.abspath(conf_dict['scores_db'])

    final_dict, emodel_dirs = prepare_emodels(
        conf_dict, continu, scores_db_path, n_processes)

    # Save output
    # TODO: gather all output business here?
    output_dir = conf_dict['output_dir']
    tools.makedirs(output_dir)
    tools.write_json(output_dir, 'final.json', final_dict)
    tools.write_json(output_dir, 'emodel_dirs.json', emodel_dirs)


def add_parser(action):
    """Add option parser"""

    parser = action.add_parser(
        'prepare',
        help='Create and prepare database with me-combinations')
    parser.add_argument('conf_filename')
    parser.add_argument('--continu', action='store_true',
                        help='continue from previous run')
    parser.add_argument('--n_processes', help='number of processes',
                        type=int)



"""Analyse scores"""

"""
Copyright (c) 2018, EPFL/Blue Brain Project

 This file is part of BluePyMM <https://github.com/BlueBrain/BluePyMM>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


# pylint: disable=R0914, C0325, W0640
# pylama: ignore=E402

import os

from bluepymm import tools

from . import sqlite_io, reporting, megate_output
from . import process_megate_config as proc_config


def select_combos(conf_filename, n_processes):
    """Parse conf file and run select combos"""
    # Parse configuration file
    conf_dict = tools.load_json(conf_filename)

    select_combos_from_conf(conf_dict, n_processes)


def select_combos_from_conf(conf_dict, n_processes=None):
    """Compare scores of me-combinations to thresholds, select successful
    combinations, and write results out to file.

    Args:
        conf_filename: filename of configuration (.json file)
        n_processes: integer number of processes, `None` will use all of them
    """
    scores_db_filename = conf_dict['scores_db']
    pdf_filename = conf_dict['pdf_filename']
    output_dir = conf_dict['output_dir']

    print('Reading configuration files')
    # read skip features
    to_skip_patterns, to_skip_features = \
        proc_config.read_to_skip_features(
            conf_dict)

    # read megate thresholds
    megate_patterns, megate_thresholds = \
        proc_config.read_megate_thresholds(
            conf_dict)

    select_perc_best = conf_dict.get('select_perc_best', None)

    print('Reading tables from sqlite')
    # read score tables
    scores, score_values = sqlite_io.read_and_process_sqlite_score_tables(
        scores_db_filename)

    print('Checking if all combos have run')
    tools.check_all_combos_have_run(scores, scores_db_filename)

    print('Start creation of ext_neurondb')
    # create final database and write report
    ext_neurondb = \
        reporting.create_final_db_and_write_report(
            pdf_filename,
            to_skip_features,
            to_skip_patterns,
            megate_thresholds,
            megate_patterns,
            conf_dict.get('skip_repaired_exemplar', False),
            conf_dict.get('check_opt_scores', True),
            scores, score_values,
            conf_dict.get('plot_emodels_per_morphology', False),
            output_dir,
            select_perc_best,
            n_processes=n_processes)
    print('Wrote pdf to %s' % os.path.abspath(pdf_filename))

    # write output files
    compliant = conf_dict.get('make_names_neuron_compliant', False)
    extneurondb_path, mecombo_emodel_path = \
        megate_output.save_megate_results(
            ext_neurondb,
            output_dir,
            sort_key='combo_name',
            make_names_neuron_compliant=compliant)

    emodels_hoc_path = conf_dict['emodels_hoc_dir']

    megate_output.write_mecomboreleasejson(
        output_dir,
        emodels_hoc_path,
        extneurondb_path,
        mecombo_emodel_path)


def add_parser(action):
    """Add parser"""

    parser = action.add_parser('select',
                               help='Select feasible me-combinations')
    parser.add_argument('conf_filename')
    parser.add_argument('--n_processes', help='number of processes',
                        type=int)



"""Tests for select_combos/reporting.py"""

from __future__ import print_function

"""
Copyright (c) 2018, EPFL/Blue Brain Project

 This file is part of BluePyMM <https://github.com/BlueBrain/BluePyMM>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


import os
import pandas

import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import pytest

from bluepymm import select_combos


BASE_DIR = os.path.abspath(os.path.dirname(__file__))
TMP_DIR = os.path.join(BASE_DIR, 'tmp/')


@pytest.mark.unit
def test_pdf_file():
    """select_combos.reporting: test pdf_file"""
    filename = 'test_report.pdf'
    path = os.path.join(TMP_DIR, filename)
    with select_combos.reporting.pdf_file(path) as pp:
        assert pp.get_pagecount() == 0
        assert os.path.exists(path)


def _get_pdf_file(filename):
    """Helper function to open PDF file."""
    path = os.path.join(TMP_DIR, filename)
    if not os.path.exists(TMP_DIR):
        os.makedirs(TMP_DIR)
    return PdfPages(path)


@pytest.mark.unit
def test_add_plot_to_report():
    """select_combos.reporting: test add_plot_to_report"""
    filename = 'test_add_plot_to_report.pdf'

    def _figure(title):
        fig = plt.figure()
        plt.title(title)
        return fig

    with _get_pdf_file(filename) as pp:
        select_combos.reporting.add_plot_to_report(pp, _figure, 'test')


@pytest.mark.unit
def test_plot_dict():
    """select_combos.reporting: test plot_dict"""
    test_dict = {'test': 1}
    title = 'test_title'
    fig = select_combos.reporting.plot_dict(test_dict, title)
    assert title == fig.get_axes()[0].get_title()
    plt.close()


@pytest.mark.unit
def test_plot_stacked_bars():
    """select_combos.reporting: test plot_stacked_bars"""
    test_data = pandas.DataFrame({'data': [1, 2, 3]})
    xlabel = 'test_xlabel'
    ylabel = 'test_ylabel'
    title = 'test_title'
    color_map = 'C0'
    fig = select_combos.reporting.plot_stacked_bars(test_data, xlabel, ylabel,
                                                    title, color_map)
    axes = fig.get_axes()[0]
    assert xlabel == axes.get_xlabel()
    assert ylabel == axes.get_ylabel()
    assert title == axes.get_title()
    plt.close()


@pytest.mark.unit
def test_plot_morphs_per_feature_for_emodel():
    """select_combos.reporting: test plot_morphs_per_feature_for_emodel"""
    emodel = 'emodel1'
    test_data = pandas.DataFrame({'passed': [True, False, True]})
    test_data_2 = pandas.DataFrame({'scores': [1, 2, 3]})
    fig = select_combos.reporting.plot_morphs_per_feature_for_emodel(
        emodel, test_data, test_data_2)
    assert emodel in fig.get_axes()[0].get_title()


@pytest.mark.unit
def test_plot_morphs_per_mtype_for_emodel():
    """select_combos.reporting: test plot_morphs_per_mtype_for_emodel"""
    emodel = 'emodel1'
    mtypes = pandas.DataFrame({'mtypes': ['mtype1', 'mtype2', 'mtype1']})
    test_scores = pandas.DataFrame({'Passed all': [True, False, True],
                                    'mtypes': ['mtype1', 'mtype2', 'mtype1']})
    fig = select_combos.reporting.plot_morphs_per_mtype_for_emodel(
        emodel, mtypes['mtypes'], test_scores)
    assert emodel in fig.get_axes()[0].get_title()


@pytest.mark.unit
def test_create_morphology_label():
    """select_combos.reporting: test create_morphology_label"""
    data = pandas.DataFrame({'morph_name': ['morph1', 'morph2'],
                             'fullmtype': ['mtype1', 'mtype2'],
                             'etype': ['etype1', 'etype2']})
    ret = select_combos.reporting.create_morphology_label(data)
    expected_ret = 'morph1 (mtype1, etype1)'
    assert ret == expected_ret


@pytest.mark.unit
def test_plot_emodels_per_morphology():
    """select_combos.reporting: test plot_emodels_per_morphology"""
    data = pandas.DataFrame({'is_exemplar': False, 'morph_name': 'morph1',
                             'exception': None, 'fullmtype': 'mtype1',
                             'etype': 'etype1'}, index=[0])
    final_db = pandas.DataFrame({'morph_name': 'morph1'}, index=[0])
    fig = select_combos.reporting.plot_emodels_per_morphology(data, final_db)
    assert 'morphology' in fig.get_axes()[0].get_title()


@pytest.mark.unit
def test_plot_emodels_per_metype():
    """select_combos.reporting: test plot_emodels_per_metype"""
    data = pandas.DataFrame({'is_exemplar': False, 'morph_name': 'morph1',
                             'exception': None, 'fullmtype': 'mtype1',
                             'etype': 'etype1'}, index=[0])
    final_db = pandas.DataFrame({'morph_name': 'morph1', 'fullmtype': 'mtype1',
                                 'etype': 'etype1'}, index=[0])
    fig = select_combos.reporting.plot_emodels_per_metype(data, final_db)
    assert 'me-type' in fig.get_axes()[0].get_title()
